/*--------------------------------------------------------------*/
/* constraints.c						*/
/*	Functions to be used on to set constaints on the layout.*/
/*--------------------------------------------------------------*/

#include "defines.h"

extern constraintptr constraints;
extern int align_penalty;		/* cost of row misalignment */
extern int ch_penalty;			/* cost function weight per channel */
extern int ch_minimum;			/* desired minimum number of channels */
extern int m1_width;
extern int m2_width;
extern int via_width;
extern int via_height;
extern int m1_space;
extern int m2_space;
extern int contact_size;

int find_net(char net_name[])
{
   int i;
   
   for (i = 0; i < netcount; i++)
      if (!strcmp(net_name, (networks + i)->name)) return i;

   fprintf(stderr, "Unknown net \"%s\" found in \"Constraints\" file.\n",
	 net_name);
   return -1;
}

void read_definitions(FILE *fi)
{
   char line_in[100], *status, *keyptr, *eqptr;
   int i, line_no;
   int weight, space;
   constraintptr newcons;
   
   ch_minimum = CHANNELS;
   ch_penalty = PENALTY;
   ch_penalty = ALIGN_PENALTY;
   m1_width = PORT_WIDTH;
   m2_width = PORT_WIDTH;
   m1_space = PORT_SPACE;
   m2_space = PORT_SPACE;
   via_width = VIA_W;
   via_height = VIA_HT;
   contact_size = CONTACT_SIZE;

   if (fi == NULL) return;	/* Defaults only */

   line_no = 0;
   line_in[0] = '\0';
   status = fgets(line_in, 99, fi);
   while (status != NULL) {
      line_no++;
      if ((eqptr = strstr(line_in, "=")) != NULL) {
	 eqptr++;
	 if (sscanf(eqptr, "%d", &weight) == 1) {
            if ((keyptr = strstr(line_in, "Channels")) != NULL) {
	       ch_minimum = weight;
            }
            else if ((keyptr = strstr(line_in, "Alignment")) != NULL) {
	       align_penalty = weight;
            }
            else if ((keyptr = strstr(line_in, "Penalty")) != NULL) {
	       ch_penalty = weight;
            }
            else if ((keyptr = strstr(line_in, "M1Width")) != NULL) {
	       m1_width = weight;
            }
            else if ((keyptr = strstr(line_in, "M2Width")) != NULL) {
	       m2_width = weight;
            }
            else if ((keyptr = strstr(line_in, "M1Space")) != NULL) {
	       m1_space = weight;
            }
            else if ((keyptr = strstr(line_in, "M2Space")) != NULL) {
	       m2_space = weight;
            }
            else if ((keyptr = strstr(line_in, "ViaWidth")) != NULL) {
	       via_width = weight;
            }
            else if ((keyptr = strstr(line_in, "ViaHeight")) != NULL) {
	       via_height = weight;
            }
            else if ((keyptr = strstr(line_in, "ContactSize")) != NULL) {
	       contact_size = weight;
            }

            /* Ignore comments and blank lines */

            else if (line_in[0] != '#' && line_in[0] != '%' && line_in[0] != '!'
			&& line_in[0] != '*' && line_in[0] != '\n') {
	       fprintf(stderr, "Invalid constraint label \"%s\" at line %d.\n"
			"Failed to read in \"Definitions\" file correctly.\n",
			line_no, line_in);
	       exit(1);
	    }
	 }
	 else
	    fprintf(stderr, "Error in \"Definitions\" file at line %d\n", line_no);
      }

      /* Grab next line */
      status = fgets(line_in, 99, fi);
   }
}

void read_constraints(FILE *fi)
{
   char line_in[100], net1[100], net2[100], conj[100], *status;
   int i, n1, n2, lines, cnstnt; // tells which constraint that is being read in
   int weight, space;
   constraintptr newcons;
   
   line_in[0] = '\0';
   status = fgets(line_in, 99, fi);
   while (status != NULL) {
      if (line_in[0] == '<') {
	 if (strstr(line_in, "Left") != NULL)
	    cnstnt = LEFT;
	 else if (strstr(line_in, "Right") != NULL)
	    cnstnt = RIGHT;
	 else if (strstr(line_in, "Top") != NULL)
	    cnstnt = TOP;
	 else if (strstr(line_in, "Bottom") != NULL)
	    cnstnt = BOTTOM;
	 else if (strstr(line_in, "Together") != NULL)
	    cnstnt = TOGETHER;
	 else {
	    fprintf(stderr, "Invalid constraint label \"%s\".\n"
		     "Failed to read in constraints file correctly.\n",
		     line_in);
	    exit(1);
	 }

	 while ((status = fgets(line_in, 99, fi)) != NULL && (line_in[0] != '<')) {
	    if ((lines = sscanf(line_in, "%s %d %d", net1, &space, &weight)) < 1)
	       continue;
	    lines--;
	    if ((n1 = find_net(net1)) != -1) {
	       newcons = (constraintptr) malloc(sizeof(constraint));
	       newcons->next = constraints;
	       constraints = newcons;
	       newcons->thisnet = n1;
	       newcons->thatnet = -1;
	       newcons->type = cnstnt;

	       if (cnstnt == TOGETHER) {
		  if ((lines = sscanf(line_in, "%*s %s %s", conj, net2)) < 1) {
		     fprintf(stderr, "Illegal line for <Together> constraint.\n");
		     exit(1);
		  }
		  else if (lines == 1) {
		     strcpy(net2, conj);
		     lines = sscanf(line_in, "%*s %*s %d %d", &space, &weight);
		  }
		  else {
		     lines = sscanf(line_in, "%*s %*s %*s %d %d", &space, &weight);
		  }
		  if ((n2 = find_net(net2)) >= 0) {
		     newcons->thatnet = n2;
		  }
		  else
		     fprintf(stderr, "Warning: unknown net "
				"\'%s\' in constraints file.\n", net2);
	       }

	       if (lines < 2) weight = 1;
	       if (lines < 1) space = 8;

	       newcons->allow = space;
	       newcons->weight = weight;
	    }
	    else
	       fprintf(stderr, "Warning: unknown net "
			"\'%s\' in constraints file.\n", net1);
	 }
	 if (status == NULL) break;
      }
   }
}
